from __future__ import annotations

import logging
from typing import Any

from airflow.providers.common.compat.sdk import BaseHook
from pyunicore import client
from pyunicore import credentials
from wtforms import StringField

logger = logging.getLogger(__name__)


class UnicoreHook(BaseHook):
    """
    Interact with Unicore.

    Creates Unicore Clients from airflow connections.

    :param uc_conn_id: The unicore connection id - default: uc_default
    """

    conn_name_attr = "uc_conn_id"
    default_conn_name = "uc_default"
    conn_type = "unicore"
    hook_name = "Unicore"

    def __init__(self, uc_conn_id: str = default_conn_name) -> None:
        super().__init__()
        self.uc_conn_id = uc_conn_id
        self.get_connection(self.uc_conn_id)

    @classmethod
    def get_connection_form_fields(cls):
        return {"auth_token": StringField("Auth Token")}

    @classmethod
    def get_ui_field_behaviour(cls) -> dict[str, Any]:
        """Return custom UI field behaviour for UNICORE connection."""
        return {
            "hidden_fields": ["schema", "port", "extra"],
            "relabeling": {"login": "Username", "host": "Site URL"},
            "placeholder": {"auth_token": "UNICORE auth token"},
        }

    def get_conn(
        self,
        overwrite_base_url: str | None = None,
        overwrite_credential: credentials.Credential | None = None,
    ) -> client.Client:
        """Return a Unicore Client. base_url and credentials may be overwritten."""
        logger.debug(
            f"Gettig connection with id '{self.uc_conn_id}' from secrets backend. Will be modified with user input for UNICORE."
        )
        params = self.get_connection(self.uc_conn_id)
        base_url = params.host
        credential = credentials.UsernamePassword(params.login, params.password)
        auth_token = params.extra_dejson.get("auth_token", None)
        if auth_token is not None:
            credential = credentials.create_credential(token=auth_token)
        if overwrite_base_url is not None:
            base_url = overwrite_base_url
        if overwrite_credential is not None:
            credential = overwrite_credential
        if not base_url:
            raise TypeError()
        conn = client.Client(credential, base_url)
        return conn

    def get_credential(self) -> credentials.Credential:
        """Return the credential part of the connection as a Credential object."""
        params = self.get_connection(self.uc_conn_id)
        credential = credentials.UsernamePassword(params.login, params.password)
        auth_token = params.extra_dejson.get("auth_token", None)
        if auth_token is not None:
            credential = credentials.create_credential(token=auth_token)
        return credential

    def get_token(self) -> str:
        """Returns a token if present, None if not."""
        params = self.get_connection(self.uc_conn_id)
        return params.extra_dejson.get("auth_token", None)

    def get_base_url(self) -> str:
        """Return the base url of the connection."""
        params = self.get_connection(self.uc_conn_id)
        return params.host

    def test_connection(self) -> tuple[bool, str]:
        """Test the connection by sending an access_info request"""
        conn = self.get_conn()
        conn.access_info()
        return True, "Connection successfully tested"



"""
to configure for executor:
- Connection details for unicore: conn_id AIRFLOW__UNICORE_EXECUTOR__UNICORE_CONN_ID | should be defined, can be skipped if every task provides one
- location (path) of python virtualenv prepared on hpc system | AIRFLOW__UNICORE_EXECUTOR__DEFAULT_ENV | should be defined, can be skipped if every task provides one

tasks should be allowed to overwrite SITE, CREDENTIALS_*, UNICORE_CONN_ID and DEFAULT_ENV - i.e. everything but the database connection - credentials should be given as a uc_credential object via executor_config

"""

import time
from typing import Any
from typing import Dict

import pyunicore.client as uc_client
from airflow.configuration import conf
from airflow.executors.base_executor import BaseExecutor
from airflow.executors.workloads import All
from airflow.executors.workloads import ExecuteTask
from airflow.models.taskinstance import TaskInstance
from airflow.models.taskinstancekey import TaskInstanceKey
from airflow.utils.state import TaskInstanceState
from pyunicore import client
from pyunicore.credentials import Credential
from pyunicore.credentials import create_credential

from ..util.job import JobDescriptionGenerator
from ..util.job import NaiveJobDescriptionGenerator

STATE_MAPPINGS: Dict[uc_client.JobStatus, TaskInstanceState] = {
    uc_client.JobStatus.UNDEFINED: TaskInstanceState.FAILED,
    uc_client.JobStatus.READY: TaskInstanceState.QUEUED,
    uc_client.JobStatus.STAGINGIN: TaskInstanceState.QUEUED,
    uc_client.JobStatus.QUEUED: TaskInstanceState.QUEUED,
    uc_client.JobStatus.RUNNING: TaskInstanceState.RUNNING,
    uc_client.JobStatus.STAGINGOUT: TaskInstanceState.RUNNING,
    uc_client.JobStatus.SUCCESSFUL: TaskInstanceState.SUCCESS,
    uc_client.JobStatus.FAILED: TaskInstanceState.FAILED,
}


class UnicoreExecutor(BaseExecutor):

    EXECUTOR_CONFIG_UNICORE_SITE_KEY = "unicore_site"  # alternative Unicore site to run at, only required if different than connection default
    EXECUTOR_CONFIG_UNICORE_CREDENTIAL_KEY = "unicore_credential"  # alternative unicore credential to use for the job, only required if different than connection default

    serve_logs = True

    def start(self):
        self.active_jobs: Dict[TaskInstanceKey, uc_client.Job] = {}
        # TODO get job description generator class and init params from config
        self.job_descr_generator: JobDescriptionGenerator = NaiveJobDescriptionGenerator()

    def _handle_used_compute_time(self, task: TaskInstanceKey, job: uc_client.Job) -> None:
        pass

    def sync(self) -> None:
        # iterate through task collection and update task/ job status - delete if needed
        for task, job in list(self.active_jobs.items()):
            state = STATE_MAPPINGS[job.status]
            if state == TaskInstanceState.FAILED:
                self.fail(task)
                self.active_jobs.pop(task)
                self._handle_used_compute_time(task, job)
            elif state == TaskInstanceState.SUCCESS:
                self.success(task)
                self.active_jobs.pop(task)
                self._handle_used_compute_time(task, job)
            elif state == TaskInstanceState.QUEUED:
                self.running_state(task, state)

        return super().sync()

    def get_task_log(self, ti: TaskInstance, try_number: int) -> tuple[list[str], list[str]]:
        # if task is still active, no neeed to check other jobs for match
        job: uc_client.Job | None = None
        if ti.key in self.active_jobs:
            job = self.active_jobs[ti.key]
        # if job is finished, need to search all jobs (that are still present on the target system)
        else:
            job = self._get_job_from_id(ti.key)

        if not job:
            return [], []

        # TODO return filecontent for stdout and stderr from jobdirectory
        logs = job.properties["log"]  # type: ignore
        working_dir = job.working_dir
        stdout = working_dir.stat("/stdout")
        stderr = working_dir.stat("/stderr")

        lines_out = stdout.raw().read().decode("utf-8").split("\n")  # type: ignore
        lines_err = stderr.raw().read().decode("utf-8").split("\n")  # type: ignore

        return logs + lines_out + lines_err, ["Test", "Where is this shown??"]

    def _get_job_from_id(self, key: TaskInstanceKey) -> uc_client.Job | None:
        # iterate over jobs and check if key matches job name
        unicore_client = self._get_unicore_client()
        # check 50 jobs at a time to prevent server from not returning some jobs
        offset = 0
        number = 50
        job_name = self.job_descr_generator.get_job_name(key)
        while True:
            partial_job_list = unicore_client.get_jobs(offset=offset, num=number)
            if len(partial_job_list) == 0:
                return None
            offset += len(partial_job_list)
            for job in partial_job_list:
                if job.properties["name"] == job_name:  # type: ignore
                    return job

    def _get_unicore_client(self, executor_config: dict | None = {}):
        overwrite_unicore_site = executor_config.get(  # type: ignore
            UnicoreExecutor.EXECUTOR_CONFIG_UNICORE_SITE_KEY, None
        )  # task can provide a different site to run at, else default from connetion is used
        overwrite_unicore_credential = executor_config.get(  # type: ignore
            UnicoreExecutor.EXECUTOR_CONFIG_UNICORE_CREDENTIAL_KEY, None
        )  # task can provide a different credential to use, else default from connection is used
        token = conf.get("unicore.executor", "AUTH_TOKEN", fallback="")
        if overwrite_unicore_credential is not None:
            token = overwrite_unicore_credential
            self.log.debug("Using user provided token.")
        base_url = conf.get(
            "unicore.executor", "DEFAULT_URL", fallback="http://localhost:8080/DEMO-SITE/rest/core"
        )
        credential: Credential = create_credential(token=token)
        if overwrite_unicore_site is not None:
            base_url = overwrite_unicore_site
            self.log.debug("Using user provided site.")
        if not base_url:
            raise TypeError()
        self.log.debug(f"Using site: {base_url}")
        self.log.debug(f"Using credential: {credential}")
        conn = client.Client(credential, base_url)
        return conn

    def _submit_job(self, workload: ExecuteTask):
        unicore_client = self._get_unicore_client(executor_config=workload.ti.executor_config)
        job_descr = self._create_job_description(workload)
        self.log.info("Generated job description")
        self.log.debug(str(job_descr))
        job = unicore_client.new_job(job_descr)
        self.log.info("Submitted unicore job")
        self.active_jobs[workload.ti.key] = job
        return job

    def _create_job_description(self, workload: ExecuteTask) -> Dict[str, Any]:
        return self.job_descr_generator.create_job_description(workload)

    def queue_workload(self, workload: ExecuteTask | All, session):
        if not isinstance(workload, ExecuteTask):
            raise TypeError(f"Don't know how to queue workload of type {type(workload).__name__}")

        # submit job to unicore and add to active_jobs dict for task state management
        job = self._submit_job(workload)
        self.active_jobs[workload.ti.key] = job

    def end(self, heartbeat_interval=10) -> None:
        # wait for current jobs to finish, dont start any new ones
        while True:
            self.sync()
            if not self.active_jobs:
                break
            time.sleep(heartbeat_interval)

    def terminate(self):
        # terminate all jobs
        for task, job in list(self.active_jobs.items()):
            job.abort()
        self.end()



from typing import Dict
from typing import List

from airflow_unicore_integration.operators.unicore_operators import (
    UnicoreGenericOperator,
)


class UnicoreContainerOperator(UnicoreGenericOperator):
    def __init__(
        self, name: str, docker_image_url: str, command: str, options: str | None = None, **kwargs
    ):
        params: Dict[str, str | List[str]] = {"COMMAND": command, "IMAGE_URL": docker_image_url}
        if options is not None:
            params["OPTIONS"] = options
        super().__init__(
            name=name,
            application_name="CONTAINER",
            application_version="1.0",
            parameters=params,
            **kwargs,
        )



import json
import logging
from typing import Any
from typing import Dict

from airflow.configuration import conf
from airflow.executors.workloads import ExecuteTask
from airflow.models.taskinstancekey import TaskInstanceKey
from airflow.providers.git.hooks.git import GitHook

from .launch_script_content import LAUNCH_SCRIPT_CONTENT_STR

logger = logging.getLogger(__name__)


class JobDescriptionGenerator:
    """
    A generator class for generating unicore jhob descriptions that may supprot different kinds of systems and/ or environments.
    """

    EXECUTOR_CONFIG_PYTHON_ENV_KEY = "python_env"  # full path to a python virtualenv that includes airflow and all required libraries for the task (without the .../bin/activate part)
    EXECUTOR_CONFIG_RESOURCES = "Resources"  # gets added to the unicore job description
    EXECUTOR_CONFIG_ENVIRONMENT = "Environment"  # gets added to the unicore job description
    EXECUTOR_CONFIG_PARAMETERS = "Parameters"  # gets added to the unicore job description
    EXECUTOR_CONFIG_PROJECT = "Project"  # gets added to the unicore job description
    EXECUTOR_CONFIG_PRE_COMMANDS = "precommands"  # gets added to the unicore job description
    EXECUTOR_CONFIG_POST_COMMANDS = "postcommands"  # gets added to the unicore job descirption
    EXECUTOR_CONFIG_JOB_TYPE = "job_type"
    EXECUTOR_CONFIG_LOGIN_NODE = "login_node"
    EXECUTOR_CONFIG_UNICORE_CONN_KEY = (
        "unicore_connection_id"  # alternative connection id for the Unicore connection to use
    )
    EXECUTOR_CONFIG_UNICORE_SITE_KEY = "unicore_site"  # alternative Unicore site to run at, only required if different than connection default
    EXECUTOR_CONFIG_UNICORE_CREDENTIAL_KEY = "unicore_credential"  # alternative unicore credential to use for the job, only required if different than connection default

    def create_job_description(self, workload: ExecuteTask) -> Dict[str, Any]:
        raise NotImplementedError()

    def get_job_name(self, key: TaskInstanceKey) -> str:
        return f"{key.dag_id} - {key.task_id} - {key.run_id} - {key.try_number}"


class NaiveJobDescriptionGenerator(JobDescriptionGenerator):
    """
    This class generates a naive unicore job, that expects there to be a working python env containing airflow and any other required dependencies on the executing system.
    """

    GIT_DAG_BUNDLE_CLASSPATH = "airflow.providers.git.bundles.git.GitDagBundle"

    def create_job_description(self, workload: ExecuteTask) -> Dict[str, Any]:
        key: TaskInstanceKey = workload.ti.key
        executor_config = workload.ti.executor_config
        if not executor_config:
            executor_config = {}
        job_descr_dict: Dict[str, Any] = {}
        # get user config from executor_config
        user_added_env: Dict[str, str] = executor_config.get(JobDescriptionGenerator.EXECUTOR_CONFIG_ENVIRONMENT, None)  # type: ignore
        user_added_params: Dict[str, str] = executor_config.get(JobDescriptionGenerator.EXECUTOR_CONFIG_PARAMETERS, None)  # type: ignore
        user_added_project: str = executor_config.get(JobDescriptionGenerator.EXECUTOR_CONFIG_PROJECT, None)  # type: ignore
        user_added_resources: Dict[str, str] = executor_config.get(JobDescriptionGenerator.EXECUTOR_CONFIG_RESOURCES, None)  # type: ignore
        user_added_pre_commands: list[str] = executor_config.get(JobDescriptionGenerator.EXECUTOR_CONFIG_PRE_COMMANDS, [])  # type: ignore
        user_defined_python_env: str = workload.ti.executor_config.get(JobDescriptionGenerator.EXECUTOR_CONFIG_PYTHON_ENV_KEY, None)  # type: ignore
        user_added_post_commands: list[str] = executor_config.get(JobDescriptionGenerator.EXECUTOR_CONFIG_POST_COMMANDS, [])  # type: ignore
        user_defined_job_type: str = executor_config.get(JobDescriptionGenerator.EXECUTOR_CONFIG_JOB_TYPE, None)  # type: ignore
        user_defined_login_node: str = executor_config.get(JobDescriptionGenerator.EXECUTOR_CONFIG_LOGIN_NODE, None)  # type: ignore
        # get local dag path from cmd and fix dag path in arguments
        dag_rel_path = str(workload.dag_rel_path)
        if dag_rel_path.startswith("DAG_FOLDER"):
            dag_rel_path = dag_rel_path[10:]
        # local_dag_path = conf.get("core", "DAGS_FOLDER") + "/" + dag_rel_path
        base_url = conf.get("api", "base_url", fallback="/")
        default_execution_api_server = f"{base_url.rstrip('/')}/execution/"
        server = conf.get(
            "unicore.executor", "execution_api_server_url", fallback=default_execution_api_server
        )
        logger.debug(f"Server is {server}")

        # set job type
        if user_defined_job_type:
            job_descr_dict["Job type"] = user_defined_job_type
        if user_defined_login_node:
            job_descr_dict["Login Node"] = user_defined_login_node

        # check which python virtualenv to use
        if user_defined_python_env:
            python_env = user_defined_python_env
        else:
            python_env = conf.get("unicore.executor", "DEFAULT_ENV")
        tmp_dir = conf.get("unicore.executor", "TMP_DIR", "/tmp")
        # prepare dag file to be uploaded via unicore
        # dag_file = open("/tmp/test")
        # dag_content = dag_file.readlines()
        # dag_import = {"To": dag_rel_path, "Data": dag_content}
        worker_script_import = {
            "To": "run_task_via_supervisor.py",
            # "From": "https://gist.githubusercontent.com/cboettcher/3f1101a1d1b67e7944d17c02ecd69930/raw/1d90bf38199d8c0adf47a79c8840c3e3ddf57462/run_task_via_supervisor.py",
            "Data": LAUNCH_SCRIPT_CONTENT_STR,
        }
        # start filling the actual job description
        job_descr_dict["Name"] = self.get_job_name(key)
        job_descr_dict["Executable"] = (
            f". airflow_config.env && . {python_env} && python run_task_via_supervisor.py --json-string '{workload.model_dump_json()}'"  # TODO may require module load to be setup for some systems
        )
        # job_descr_dict["Arguments"] = [
        #    "-c",
        #    "source airflow_config.env",
        #    "source {python_env}/bin/activate",
        #    "python",
        #    "run_task_via_supervisor.py",
        #    f"--json-string '{workload.model_dump_json()}'",
        # ]

        job_descr_dict["Environment"] = {
            "AIRFLOW__CORE__EXECUTION_API_SERVER_URL": server,
            # "AIRFLOW__CORE__DAGS_FOLDER": "./",
            "AIRFLOW__LOGGING__LOGGING_LEVEL": "DEBUG",
            "AIRFLOW__CORE__EXECUTOR": "LocalExecutor,airflow_unicore_integration.executors.unicore_executor.UnicoreExecutor",
        }

        # build filecontent string for importing in the job | this is needed to avoid confusing nested quotes and trying to escape them properly when using unicore env vars directly
        env_file_content: list[str] = []

        # transmit needed dag bundle information (and possibly files) to job directory
        bundle_str = conf.get("dag.processor", "dag_bundle_config_list")
        logger.debug(f"Dag Bundle config is: {bundle_str}")
        bundle_dict = json.loads(bundle_str)
        conn_id_to_transmit = None
        bundle_type = None

        for bundle in bundle_dict:
            if bundle["name"] == workload.bundle_info.name:
                if bundle["classpath"] == NaiveJobDescriptionGenerator.GIT_DAG_BUNDLE_CLASSPATH:
                    bundle_type = NaiveJobDescriptionGenerator.GIT_DAG_BUNDLE_CLASSPATH
                    env_file_content.append(
                        f"export AIRFLOW__DAG_PROCESSOR__DAG_BUNDLE_CONFIG_LIST='[{json.dumps(bundle)}]'"
                    )
                    conn_id_to_transmit = bundle["kwargs"]["git_conn_id"]
                    break
                # TODO handle other bundle types

        if bundle_type:
            if (
                bundle_type == NaiveJobDescriptionGenerator.GIT_DAG_BUNDLE_CLASSPATH
                and conn_id_to_transmit
            ):
                git_hook = GitHook(conn_id_to_transmit)
                git_remote_url = git_hook.repo_url
                git_dir_prefix = f"{tmp_dir}/{workload.ti.dag_id}/{workload.ti.task_id}/{workload.ti.run_id}/{workload.ti.try_number}"
                git_local_url = f"{git_dir_prefix}/dagmirror"
                dag_bundle_path = f"{git_dir_prefix}/dagbundle"
                # add precommand to clone repo on login node
                git_precommand = f". {python_env} && mkdir -p {git_local_url} && mkdir -p {dag_bundle_path} && git clone {git_remote_url} {git_local_url}"
                job_descr_dict["Environment"][
                    "AIRFLOW__DAG_PROCESSOR__DAG_BUNDLE_STORAGE_PATH"
                ] = f"{dag_bundle_path}"
                logger.info(f"git precommand is {git_precommand}")
                user_added_pre_commands.append(git_precommand)
                # add connection to local clone to env of job
                airflow_conn_string = json.dumps(
                    {"conn_type": "git", "host": f"file://{git_local_url}"}
                )
                env_file_content.append(
                    f"export AIRFLOW_CONN_{str(conn_id_to_transmit).upper()}='{airflow_conn_string}'"
                )
                logger.info(f"connection is '{airflow_conn_string}'")
                # add cleanup of local git repo to job description
                git_cleanup_command = f"rm -r {git_dir_prefix}"
                logger.info(f"git cleanup is {git_cleanup_command}")
                user_added_post_commands.append(git_cleanup_command)

        airflow_env_import = {"To": "airflow_config.env", "Data": env_file_content}

        job_descr_dict["Imports"] = [worker_script_import, airflow_env_import]

        if len(user_added_pre_commands) > 0:
            precommand_import = {"To": "precommand.sh", "Data": user_added_pre_commands}
            job_descr_dict["Imports"].append(precommand_import)
            job_descr_dict["User precommand"] = "bash precommand.sh"
        if len(user_added_post_commands) > 0:
            postcommand_import = {"To": "postcommand.sh", "Data": user_added_post_commands}
            job_descr_dict["Imports"].append(postcommand_import)
            job_descr_dict["User postcommand"] = "bash postcommand.sh"

        job_descr_dict["RunUserPrecommandOnLoginNode"] = (
            "true"  # precommand needs public internet access to clone dag repos
        )
        # add user defined options to description
        if user_added_env:
            job_descr_dict["Environment"].update(user_added_env)
        if user_added_params:
            job_descr_dict["Parameters"] = user_added_params
        if user_added_project:
            job_descr_dict["Project"] = user_added_project
        if user_added_resources:
            job_descr_dict["Resources"] = user_added_resources

        return job_descr_dict
